@@ -51,7 +51,17 @@ pub fn new<'i, R: RuleType>(
 ) -> Pairs<'i, R> {
     let line_index = match line_index {
         Some(line_index) => line_index,
-        None => Rc::new(LineIndex::new(input)),
+        None => {
+            let last_input_pos = queue
+                .last()
+                .map(|token| match *token {
+                    QueueableToken::Start { input_pos, .. }
+                    | QueueableToken::End { input_pos, .. } => input_pos,
+                })
+                .unwrap_or(0);
+
+            Rc::new(LineIndex::new(&input[..last_input_pos]))
+        }
     };
 
     let mut pairs_count = 0;
@@ -205,7 +215,13 @@ impl<'i, R: RuleType> Pairs<'i, R> {
     /// ```
     #[inline]
     pub fn flatten(self) -> FlatPairs<'i, R> {
-        unsafe { flat_pairs::new(self.queue, self.input, self.start, self.end) }
+        flat_pairs::new(
+            self.queue,
+            self.input,
+            self.line_index,
+            self.start,
+            self.end,
+        )
     }
 
     /// Finds the first pair that has its node or branch tagged with the provided
@@ -227,9 +243,9 @@ impl<'i, R: RuleType> Pairs<'i, R> {
     ///     state: Box<ParserState<'_, Rule>>,
     /// ) -> ParseResult<Box<ParserState<'_, Rule>>> {
     ///     expr(state, Rule::mul, "*")
-    ///         .and_then(|state| state.tag_node(std::borrow::Cow::Borrowed("mul")))
+    ///         .and_then(|state| state.tag_node("mul"))
     ///         .or_else(|state| expr(state, Rule::add, "+"))
-    ///         .and_then(|state| state.tag_node(std::borrow::Cow::Borrowed("add")))
+    ///         .and_then(|state| state.tag_node("add"))
     /// }
     /// fn expr<'a>(
     ///     state: Box<ParserState<'a, Rule>>,
@@ -239,10 +255,10 @@ impl<'i, R: RuleType> Pairs<'i, R> {
     ///     state.rule(r, |state| {
     ///         state.sequence(|state| {
     ///             number(state)
-    ///                 .and_then(|state| state.tag_node(std::borrow::Cow::Borrowed("lhs")))
+    ///                 .and_then(|state| state.tag_node("lhs"))
     ///                 .and_then(|state| state.match_string(o))
     ///                 .and_then(number)
-    ///                 .and_then(|state| state.tag_node(std::borrow::Cow::Borrowed("rhs")))
+    ///                 .and_then(|state| state.tag_node("rhs"))
     ///         })
     ///     })
     /// }
@@ -278,9 +294,9 @@ impl<'i, R: RuleType> Pairs<'i, R> {
     ///     state: Box<ParserState<'_, Rule>>,
     /// ) -> ParseResult<Box<ParserState<'_, Rule>>> {
     ///     expr(state, Rule::mul, "*")
-    ///         .and_then(|state| state.tag_node(std::borrow::Cow::Borrowed("mul")))
+    ///         .and_then(|state| state.tag_node("mul"))
     ///         .or_else(|state| expr(state, Rule::add, "+"))
-    ///         .and_then(|state| state.tag_node(std::borrow::Cow::Borrowed("add")))
+    ///         .and_then(|state| state.tag_node("add"))
     /// }
     /// fn expr<'a>(
     ///     state: Box<ParserState<'a, Rule>>,
@@ -290,10 +306,10 @@ impl<'i, R: RuleType> Pairs<'i, R> {
     ///     state.rule(r, |state| {
     ///         state.sequence(|state| {
     ///             number(state)
-    ///                 .and_then(|state| state.tag_node(std::borrow::Cow::Borrowed("lhs")))
+    ///                 .and_then(|state| state.tag_node("lhs"))
     ///                 .and_then(|state| state.match_string(o))
     ///                 .and_then(number)
-    ///                 .and_then(|state| state.tag_node(std::borrow::Cow::Borrowed("rhs")))
+    ///                 .and_then(|state| state.tag_node("rhs"))
     ///         })
     ///     })
     /// }
@@ -311,7 +327,7 @@ impl<'i, R: RuleType> Pairs<'i, R> {
     pub fn find_tagged(
         self,
         tag: &'i str,
-    ) -> Filter<FlatPairs<'i, R>, impl FnMut(&Pair<'i, R>) -> bool + '_> {
+    ) -> Filter<FlatPairs<'i, R>, impl FnMut(&Pair<'i, R>) -> bool + 'i> {
         self.flatten()
             .filter(move |pair: &Pair<'i, R>| matches!(pair.as_node_tag(), Some(nt) if nt == tag))
     }
@@ -347,19 +363,22 @@ impl<'i, R: RuleType> Pairs<'i, R> {
     #[inline]
     pub fn peek(&self) -> Option<Pair<'i, R>> {
         if self.start < self.end {
-            Some(unsafe {
-                pair::new(
-                    Rc::clone(&self.queue),
-                    self.input,
-                    Rc::clone(&self.line_index),
-                    self.start,
-                )
-            })
+            Some(pair::new(
+                Rc::clone(&self.queue),
+                self.input,
+                Rc::clone(&self.line_index),
+                self.start,
+            ))
         } else {
             None
         }
     }
 
+    /// Returns `true` if the iterator contains no `Pair`s.
+    pub fn is_empty(&self) -> bool {
+        self.pairs_count == 0
+    }
+
     /// Generates a string that stores the lexical information of `self` in
     /// a pretty-printed JSON format.
     #[cfg(feature = "pretty-print")]
@@ -394,7 +413,7 @@ impl<'i, R: RuleType> Pairs<'i, R> {
     }
 }
 
-impl<'i, R: RuleType> ExactSizeIterator for Pairs<'i, R> {
+impl<R: RuleType> ExactSizeIterator for Pairs<'_, R> {
     #[inline]
     fn len(&self) -> usize {
         self.pairs_count
@@ -418,7 +437,7 @@ impl<'i, R: RuleType> Iterator for Pairs<'i, R> {
     }
 }
 
-impl<'i, R: RuleType> DoubleEndedIterator for Pairs<'i, R> {
+impl<R: RuleType> DoubleEndedIterator for Pairs<'_, R> {
     fn next_back(&mut self) -> Option<Self::Item> {
         if self.end <= self.start {
             return None;
@@ -427,26 +446,24 @@ impl<'i, R: RuleType> DoubleEndedIterator for Pairs<'i, R> {
         self.end = self.pair_from_end();
         self.pairs_count -= 1;
 
-        let pair = unsafe {
-            pair::new(
-                Rc::clone(&self.queue),
-                self.input,
-                Rc::clone(&self.line_index),
-                self.end,
-            )
-        };
+        let pair = pair::new(
+            Rc::clone(&self.queue),
+            self.input,
+            Rc::clone(&self.line_index),
+            self.end,
+        );
 
         Some(pair)
     }
 }
 
-impl<'i, R: RuleType> fmt::Debug for Pairs<'i, R> {
+impl<R: RuleType> fmt::Debug for Pairs<'_, R> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_list().entries(self.clone()).finish()
     }
 }
 
-impl<'i, R: RuleType> fmt::Display for Pairs<'i, R> {
+impl<R: RuleType> fmt::Display for Pairs<'_, R> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         write!(
             f,
@@ -468,7 +485,7 @@ impl<'i, R: PartialEq> PartialEq for Pairs<'i, R> {
     }
 }
 
-impl<'i, R: Eq> Eq for Pairs<'i, R> {}
+impl<R: Eq> Eq for Pairs<'_, R> {}
 
 impl<'i, R: Hash> Hash for Pairs<'i, R> {
     fn hash<H: Hasher>(&self, state: &mut H) {
@@ -480,7 +497,7 @@ impl<'i, R: Hash> Hash for Pairs<'i, R> {
 }
 
 #[cfg(feature = "pretty-print")]
-impl<'i, R: RuleType> ::serde::Serialize for Pairs<'i, R> {
+impl<R: RuleType> ::serde::Serialize for Pairs<'_, R> {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
         S: ::serde::Serializer,
@@ -676,9 +693,9 @@ mod tests {
             state: Box<ParserState<'_, Rule>>,
         ) -> ParseResult<Box<ParserState<'_, Rule>>> {
             expr(state, Rule::mul, "*")
-                .and_then(|state| state.tag_node(alloc::borrow::Cow::Borrowed("mul")))
+                .and_then(|state| state.tag_node("mul"))
                 .or_else(|state| expr(state, Rule::add, "+"))
-                .and_then(|state| state.tag_node(alloc::borrow::Cow::Borrowed("add")))
+                .and_then(|state| state.tag_node("add"))
         }
         fn expr<'a>(
             state: Box<ParserState<'a, Rule>>,
@@ -688,10 +705,10 @@ mod tests {
             state.rule(r, |state| {
                 state.sequence(|state| {
                     number(state)
-                        .and_then(|state| state.tag_node(alloc::borrow::Cow::Borrowed("lhs")))
+                        .and_then(|state| state.tag_node("lhs"))
                         .and_then(|state| state.match_string(o))
                         .and_then(number)
-                        .and_then(|state| state.tag_node(alloc::borrow::Cow::Borrowed("rhs")))
+                        .and_then(|state| state.tag_node("rhs"))
                 })
             })
         }
