@@ -24,16 +24,12 @@ use crate::RuleType;
 /// [`Pairs::tokens`]: struct.Pairs.html#method.tokens
 #[derive(Clone)]
 pub struct Tokens<'i, R> {
-    /// # Safety:
-    ///
-    /// All `QueueableToken`s' `input_pos` must be valid character boundary indices into `input`.
     queue: Rc<Vec<QueueableToken<'i, R>>>,
     input: &'i str,
     start: usize,
     end: usize,
 }
 
-// TODO(safety): QueueableTokens must be valid indices into input.
 pub fn new<'i, R: RuleType>(
     queue: Rc<Vec<QueueableToken<'i, R>>>,
     input: &'i str,
@@ -46,7 +42,7 @@ pub fn new<'i, R: RuleType>(
                 QueueableToken::Start { input_pos, .. } | QueueableToken::End { input_pos, .. } => {
                     assert!(
                         input.get(input_pos..).is_some(),
-                        "💥 UNSAFE `Tokens` CREATED 💥"
+                        "💥 INVALID `Tokens` CREATED 💥"
                     )
                 }
             }
@@ -75,24 +71,20 @@ impl<'i, R: RuleType> Tokens<'i, R> {
 
                 Token::Start {
                     rule,
-                    // QueueableTokens are safely created.
-                    pos: unsafe { position::Position::new_unchecked(self.input, input_pos) },
+                    pos: position::Position::new_internal(self.input, input_pos),
                 }
             }
             QueueableToken::End {
                 rule, input_pos, ..
-            } => {
-                Token::End {
-                    rule,
-                    // QueueableTokens are safely created.
-                    pos: unsafe { position::Position::new_unchecked(self.input, input_pos) },
-                }
-            }
+            } => Token::End {
+                rule,
+                pos: position::Position::new_internal(self.input, input_pos),
+            },
         }
     }
 }
 
-impl<'i, R: RuleType> ExactSizeIterator for Tokens<'i, R> {
+impl<R: RuleType> ExactSizeIterator for Tokens<'_, R> {
     fn len(&self) -> usize {
         self.end - self.start
     }
@@ -119,7 +111,7 @@ impl<'i, R: RuleType> Iterator for Tokens<'i, R> {
     }
 }
 
-impl<'i, R: RuleType> DoubleEndedIterator for Tokens<'i, R> {
+impl<R: RuleType> DoubleEndedIterator for Tokens<'_, R> {
     fn next_back(&mut self) -> Option<Self::Item> {
         if self.end <= self.start {
             return None;
@@ -133,7 +125,7 @@ impl<'i, R: RuleType> DoubleEndedIterator for Tokens<'i, R> {
     }
 }
 
-impl<'i, R: RuleType> fmt::Debug for Tokens<'i, R> {
+impl<R: RuleType> fmt::Debug for Tokens<'_, R> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_list().entries(self.clone()).finish()
     }
